{
  TExtendedEdit
  -------------
  Copyright  2000-2002 FVAL - Consultoria e Informtica Ltda.


  TCustomMaskEdit heritage. Has capability to align contents and input of
  numeric, date e time values.

  Author  : Fernando Val
  email   : fernando@fval.com.br
  Version : 3.20
  Date    : Jan/16/2002
  License : FreeWare for all purposes

  Based on:
    TEditNew v1.03 - Jos M Gas <sigecom@bbvnet.es>
    TAlignEdit - autor desconhecido (unknown author)
    TNumericEdit - Rafael Bulkool

  Informations
  ------------
  This component was and tested on Delphi 5.

  3.20 - Jan/16/2002 - Fernando Val <fernando@fval.com.br>

       - Added TimeFormat property to determine the format of time entry mask.
         The possible values are: tfHMS = HH:NN:SS; tfHM  = HH:NN; tfMS  = NN:SS;
       - Added MaxValue and MinValue properties.

  3.10 - Jan/15/2002 - Fernando Val <fernando@fval.com.br>

       - Added InvalidDateMessage and InvalidTimeMessage properties.
         %text% is a macro for Text property.

  3.00 - Jan/14/2002 - Fernando Val <fernando@fval.com.br>

       - Hierarchy changed from TEdit to TCustomMaskEdit;
       - Added ValueDateTime (TDateTime) property;
       - Implemented inputs for dates e times with optional checkup
         (CheckOnExit property).

  2.01 - Aug/02/2001 - Fernando Val <fernando@fval.com.br>

       + Fixed the Enter and Esc key bug:
         Now the Enter key call default button click if any exists and the
         Escape key restores initial value of text property if it was changed
         other wise call form's default escape method.

  2.00 - Jul/25/2001 - Fernando Val <fernando@fval.com.br>

       - Appended the right alignment feature while editing;
       - Fixed some bugs;
       + Added the following properties:
         AutoFormatDecimal -
         Decimal  - Defines the decimal separator
         Thousand - Defines the char for thousand separator


  1.10 - Jun/11/2001 - Fernando Val <fernando@fval.com.br>

       + Added the following propertie:
         UseThousandSeparator - Determines de use of thousand separator


  1.00 - Jul/27/2000 - Fernando Val <fval@fval.com.br>, Jos M Gas <sigecom@bbvnet.es>

       - Joined the two components in one;
       - Imported the following properties:
         ColorOnFocus    - Color de fondo cuando el Edit tiene el Foco
         ColorOnNotFocus - Color de fondo cuando el Edit No tiene el Foco
         EditPrecision   - Permite definir la cantidad de decimales. Si el valor es
                           mayor de 0, formatear el valor, completando con ceros al
                           final si no hay decimales introducidos.

                           ## En el caso de ser valor Float, el separador decimal lo
                           coge por defecto de Windows.
                           Defines the amount of decimal digits. If value is greater
                           then 0, format value completing with zeros after

         EditType        - etString, etInteger, etFloat. Consideracin de los datos
                           que contendr el Edit
         FirstCharUpper  - Convierte a mayusculas el primer Caracter de la cadena si
                           EditType = etString
         KeyByTab        - Respuesta al pulsar una tecla como si se pulsara Tabulador
                           para salir del componente, pasando el foco al siguiente, con
                           el valor de otra tecla.

                           Por defecto se asigna el estandar de Windows Tabulador, para
                           anularlo: KeyByTab = #9

                           Por ejemplo, si deseamos que de un TEditN se pueda pasar el
                           foco al siguiente objeto con la tecla Enter ademas de con
                           el Tabulador, pondriamos esta valor a KeyByTab = #13

         ValueInteger    - Integer value of TExtendedEdit, when EditType = etInteger
                           If occurs an conversion error then result 0
         ValueFloat      - Float values of TExtendedEdit, when EditType = etFloat
                           If occurs an conversion error then result 0
         WidthOnFocus    - Permite definir una anchura del TEditN diferente cuando el
                           EditN tiene el Foco. La anchura no se ajusta en funcin del
                           texto introducido, debiendose definirse en modo diseo.
}

{$IFNDEF VER80}              { Delphi 1.0     }
 {$IFNDEF VER90}             { Delphi 2.0     }
  {$IFNDEF VER93}            { C++Builder 1.0 }
    {$IFNDEF VER100}         { Delphi 3.0     }
      {$IFNDEF VER110}       { C++Builder 3.0 }
        {$IFNDEF VER120}     { Delphi 4.0     }
          {$IFNDEF VER125}   { C++Builder 4.0 }
            {$DEFINE DELFI5} { Delphi 5.0 or higher }
          {$ENDIF}
        {$ENDIF}
      {$ENDIF}
    {$ENDIF}
  {$ENDIF}
 {$ENDIF}
{$ENDIF}

unit ExtendedEdit;

{$R-}

interface

uses
  {$IFDEF WIN32}Windows,{$ELSE}Winprocs,{$ENDIF}
  Messages, SysUtils, Classes, StdCtrls, Controls, Graphics, Forms, Dialogs, Mask;

type
  TDateOrder = (doDMY, doMDY, doYMD);
  TTimeFormat = (tfHMS, tfHM, tfMS);
  TEditTypes = (etString, etInteger, etFloat, etDate, etTime);
  TSeparator = (spComma, spPoint);

  TCustomExtendedEdit = class(TCustomMaskEdit)
  private
    FAlignment: TAlignment;
    FCanvas: TControlCanvas;
    FFocused: Boolean;
    procedure SetFocused(Value: Boolean);
    //function GetTextMargins: TPoint;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    //procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
  protected
    procedure SetAlignment(Value: TAlignment);
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
  public
    destructor Destroy; override;
  end;

  TExtendedEdit = class(TCustomExtendedEdit)
  private
    FAutoFormatDecimal: Boolean;
    FChangeThousand   : Boolean;
    FCheckOnExit      : Boolean;
    FColorOnFocus     : TColor;
    FColorOnNoFocus   : TColor;
    FDateOrder        : TDateOrder;
    FDateSeparator    : Char;
    FDateTimeMask     : String;
    FDecimal          : TSeparator;
    FFontOnFocus      : TFont;
    FFontOnNoFocus    : TFont;
    FExitOnReturn     : Boolean;
    FEditType         : TEditTypes;
    FKeyTab           : Char;
    FMaxValue         : Variant;
    FMinValue         : Variant;
    FMsgInvDate       : String;
    FMsgInvTime       : String;
    FPrecision        : Integer;
    FSDecimal         : Char;
    FSThousand        : Char;
    FTimeFormat       : TTimeFormat;
    FThousand         : TSeparator;
    FThousandCount    : Word;
    FUseThousand      : Boolean;
    FUpper            : Boolean;
    FValDateTime      : TDateTime;
    FValInteger       : Integer;
    FValFloat         : Double;
    FWidthOnFocus     : Integer;
    iWidth            : Integer;
    TextAtEnter       : String;
    TxtConvert        : String;
    ValTemp           : Extended;
    FOnChange         : TNotifyEvent;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure ChangeSeparators;
    procedure FormatThousand;
    procedure InitializeValue;
    procedure ValidateDigit(var Key: Char);
    procedure SetAutoFormatDecimal(Value: Boolean);
    procedure SetColorOnFocus(Value: TColor);
    procedure SetColorOnNoFocus(Value: TColor);
    procedure SetDecimal(Value: TSeparator);
    procedure SetFontOnFocus(Value: TFont);
    procedure SetFontOnNoFocus(Value: TFont);
    procedure SetEditType(Value: TEditTypes);
    procedure SetInteger(Value: Integer);
    procedure SetFloat(Value: Double);
    procedure SetDateTime(Value: TDateTime);
    procedure SetPrecision(Value: Integer);
    procedure SetDateOrder(Value: TDateOrder);
    procedure SetTimeFormat(Value: TTimeFormat);
    procedure SetThousand(Value: TSeparator);
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure FontChanged(Sender: TObject);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override; // ExitOnReturn
    procedure KeyPress(var Key: Char); override;                    // ExitOnReturn
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure SetDateEditMask;
    procedure CheckDateTime;
  protected
    { Protected declarations }
  published
    property Alignment;
{IFDEF DELFI5}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ImeMode;
    property ImeName;
    property ParentBiDiMode;
{ENDIF}
    property AutoFormatDecimal: Boolean read FAutoFormatDecimal write SetAutoFormatDecimal default False;
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property CharCase;
    property CheckOnExit: Boolean read FCheckOnExit write FCheckOnExit;
    property ColorOnFocus: TColor read FColorOnFocus write SetColorOnFocus;
    property ColorOnNoFocus: TColor read FColorOnNoFocus write SetColorOnNoFocus;
    property Ctl3D;
    property DateOrder: TDateOrder read FDateOrder write SetDateOrder;
    property DateSeparator: Char read FDateSeparator;
    property Decimal: TSeparator read FDecimal write SetDecimal;
    property DragCursor;
    property DragMode;
    property Enabled;
    property EditPrecision: Integer read FPrecision write SetPrecision;
    property EditType: TEditTypes read FEditType write SetEditType;
    property EditKeyByTab: Char read FKeyTab write FKeyTab;
    property ExitOnReturn: boolean read FExitOnReturn write FExitOnReturn default false;
    property FirstCharUpper: Boolean read FUpper write FUpper;
    property FontOnFocus: TFont read FFontOnFocus write SetFontOnFocus;
    property FontOnNoFocus: TFont read FFontOnNoFocus write SetFontOnNoFocus;
    property HideSelection;
    property InvalidDateMessage: String read FMsgInvDate write FMsgInvDate;
    property InvalidTimeMessage: String read FMsgInvTime write FMsgInvTime;
    property MaxLength;
    property MaxValue: Variant read FMaxValue write FMaxValue;
    property MinValue: Variant read FMinValue write FMinValue;
    property OEMConvert;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Thousand: TSeparator read FThousand write SetThousand;
    property TimeFormat: TTimeFormat read FTimeFormat write SetTimeFormat;
    property UseThousandSeparator: Boolean read FUseThousand write FUseThousand;
    property ValueDateTime: TDateTime read FValDateTime write SetDateTime;
    property ValueInteger: Integer read FValInteger write SetInteger;
    property ValueFloat: Double read FValFloat write SetFloat;
    property Visible;
    property WidthOnFocus: Integer read FWidthOnFocus write FWidthOnFocus;
    property OnChange: TNotifyEvent read FOnChange  write FOnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  public
    property Font;
    property Color;
    procedure Change; override;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

const
  comma  = ',';
  point  = '.';
  minus  = '-';
  plus   = '+';
  back   = Char(VK_BACK);
  EmptyDate : TDateTime = 0;

procedure Register;

implementation

{ TCustomExtendedEdit }

procedure TCustomExtendedEdit.SetAlignment(Value: TAlignment);
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
//    Invalidate;
    RecreateWnd;  // Este mtodo permite, destruindo e criando a janela interna que o contedo permanea no alinhamento correto
  end;
end;

procedure TCustomExtendedEdit.SetFocused(Value: Boolean);
begin
  if FFocused <> Value then
  begin
    FFocused := Value;
    if (FAlignment <> taLeftJustify) then
      Invalidate;
  end;
end;

procedure TCustomExtendedEdit.CMEnter(var Message: TCMEnter);
begin
  Self.SelStart  := 0;
  Self.SelLength := Length(Self.text);
  SetFocused(True);
  inherited;
end;

destructor TCustomExtendedEdit.Destroy;
begin
  FCanvas.Free;
  inherited Destroy;
end;

{
  The following e procedures was removed on version 2.0 because they are not more necessary.

procedure TCustomExtendedEdit.WMPaint(var Message: TWMPaint);
var
  Left: Integer;
  Margins: TPoint;
  R: TRect;
  DC: HDC;
  PS: TPaintStruct;
  S: string;
begin
  if ((FAlignment = taLeftJustify) or FFocused) and
     not (csPaintCopy in ControlState) then
  begin
    inherited;
    exit;
  end;
  // Since edit controls do not handle justification unless multi-line (and
  // then only poorly) we will draw right and center justify manually unless
  // the edit has the focus.
  if FCanvas = nil then
  begin
    FCanvas := TControlCanvas.Create;
    FCanvas.Control := Self;
  end;
  DC := Message.DC;
  if DC = 0 then
    DC := BeginPaint(Handle, PS);
  FCanvas.Handle := DC;
  try
    FCanvas.Font := Font;
    if not Enabled then
      FCanvas.Font.Color := clGrayText;
    with FCanvas do
    begin
      R := ClientRect;
      if not (NewStyleControls and Ctl3D) and (BorderStyle = bsSingle) then
      begin
        Brush.Color := clWindowFrame;
        FrameRect(R);
        InflateRect(R, -1, -1);
      end;
      Brush.Color := Color;
      if (csPaintCopy in ControlState) then
        S := ''
      else
        S := Text;
      Margins := GetTextMargins;
      case FAlignment of
        taLeftJustify: Left := Margins.X;
        taRightJustify: Left := ClientWidth - TextWidth(S) - Margins.X - 1;
      else
        Left := (ClientWidth - TextWidth(S)) div 2;
      end;
      TextRect(R, Left, Margins.Y, S);
    end;
  finally
    FCanvas.Handle := 0;
    if Message.DC = 0 then
      EndPaint(Handle, PS);
  end;
end;

function TCustomExtendedEdit.GetTextMargins: TPoint;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  if NewStyleControls then
  begin
    if BorderStyle = bsNone then
      I := 0
    else
      if Ctl3D then
        I := 1
      else
        I := 2;
    Result.X := SendMessage(Handle, EM_GETMARGINS, 0, 0) and $0000FFFF + I;
    Result.Y := I;
  end
  else
  begin
    if BorderStyle = bsNone then
      I := 0
    else
    begin
      DC := GetDC(0);
      GetTextMetrics(DC, SysMetrics);
      SaveFont := SelectObject(DC, Font.Handle);
      GetTextMetrics(DC, Metrics);
      SelectObject(DC, SaveFont);
      ReleaseDC(0, DC);
      I := SysMetrics.tmHeight;
      if I > Metrics.tmHeight then
        I := Metrics.tmHeight;
      I := I div 4;
    end;
    Result.X := I;
    Result.Y := I;
  end;
end;}

procedure TCustomExtendedEdit.CMExit(var Message: TCMExit);
begin
  SetFocused(False);
  DoExit;
end;

{ TExtendedEdit}

constructor TExtendedEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FColorOnFocus   := clWindow;
  FColorOnNoFocus := clWindow;
  FFontOnFocus    := TFont.Create;
  FFontOnNoFocus  := TFont.Create;
  FFontOnNoFocus.OnChange := FontChanged;

  Color              := FColorOnNoFocus;
  FAutoFormatDecimal := False;
  FCheckOnExit       := True;
  FDateOrder         := doDMY;
  FDateSeparator     := '/';
  FDateTimeMask      := 'DD/MM/YYYY HH:NN:SS';
  FEditType          := etString;
  FKeyTab            := #9;
  FMaxValue          := 0;
  FMinValue          := 0;
  FMsgInvDate        := '"%text%" is an invalid date.';
  FMsgInvTime        := '"%text%" is an invalid time.';
  FValDateTime       := EmptyDate;
  FValInteger        := 0;
  FValFloat          := 0;
  FPrecision         := 0;
  FSDecimal          := DecimalSeparator;
  FSThousand         := ThousandSeparator;
  FThousandCount     := 3;
  FChangeThousand    := False;
  FUseThousand       := True;
  FUpper             := False;
  FWidthOnFocus      := 0;
  TextAtEnter        := '';
  EditMask           := '';

  case DecimalSeparator of
    comma: FDecimal := spComma;
    point: FDecimal := spPoint;
  end;
  case ThousandSeparator of
    comma: FThousand := spComma;
    point: FThousand := spPoint;
  end;
  InitializeValue;
end;

procedure TExtendedEdit.CreateParams(var Params: TCreateParams);
const
  Alignments: array[TAlignment] of Longint = (ES_LEFT, ES_RIGHT, ES_CENTER);
begin
  inherited CreateParams(Params);
  CreateSubClass(Params, 'EDIT');
  //Params.Style := Params.Style and not (ES_AUTOVSCROLL or ES_AUTOHSCROLL) or ES_MULTILINE or Alignments[FAlignment] ;
  Params.Style := Params.Style and not (ES_AUTOVSCROLL) or ES_MULTILINE or Alignments[FAlignment] ;
end;

destructor TExtendedEdit.Destroy;
begin
  FFontOnFocus.free;
  FFontOnNoFocus.free;
  inherited Destroy;
end;

procedure TExtendedEdit.SetAutoFormatDecimal(Value: Boolean);
begin
  if (FAutoFormatDecimal <> Value) then
    FAutoFormatDecimal := Value;
end;

procedure TExtendedEdit.SetDecimal(Value: TSeparator);
begin
  if (FDecimal <> Value) then
  begin
    //troca os separadores de milhar pelo separador decimal e vice-versa
    ChangeSeparators;
    //quando um  (,) o outro tem que ser (.)
    FThousand := FDecimal;
    FDecimal  := Value;
  end;
end;

procedure TExtendedEdit.SetThousand(Value: TSeparator);
begin
  if (FThousand <> Value) then
  begin
    //troca os separadores de milhar pelo separador decimal e vice-versa
    ChangeSeparators;
    //troca os separadores (quando um  (,) o outro tem que ser (.))
    FDecimal  := FThousand;
    FThousand := Value;
  end;
end;

procedure TExtendedEdit.ChangeSeparators;
var
  spPosition : Word;
  buffer     : String;
  i          : Integer;
begin
  buffer := Self.text;
  //busca a posio do separador decimal e troca pelo separador de milhar
  case FDecimal of
    spComma:
      begin
        spPosition := Pos(comma, Self.text);
        if (spPosition > 0) then
          buffer[spPosition] := point;
        //troca todos os separadores de milhar pelo separador decimal
        i := Pos(point, buffer);
        while (i > 0) and (i <> spPosition) do
        begin
          buffer[i] := comma;
          i := Pos(point, buffer);
        end;
      end;
    spPoint:
      begin
        spPosition := Pos(point, Self.text);
        if (spPosition > 0) then
          buffer[spPosition] := comma;
        //troca todos os separadores de milhar pelo separador decimal
        i := Pos(comma, buffer);
        while (i > 0) and (i <> spPosition) do
        begin
          buffer[i] := point;
          i := Pos(comma, buffer);
        end;
      end;
  end;
  Self.Text := buffer;
end;

procedure TExtendedEdit.InitializeValue;
var
  buffer     : String;
  i, j       : Integer;
  spDecimal,
  spThousand : Char;
  valid      : Boolean;
begin
  //busca o separador decimal
  case FDecimal of
    spComma: spDecimal := comma;
    spPoint: spDecimal := point;
  end;

  //busca o separador de milhar
  case FThousand of
    spComma: spThousand := comma;
    spPoint: spThousand := point;
  end;

  buffer := Self.Text;
  //se o texto for vazio, ento acepta o valor zero
  if (buffer = '') and (FEditType <> etString) then
  begin
    buffer := '0' + spDecimal;
    for i := 1 to FPrecision do
      buffer := buffer + '0';
  end
  else if (FEditType <> etString) then
  begin
    //se achar algum caracter invlido, ento acepta o valor zero
    valid := True;
    for i := 1 to Length(buffer) do
      if not(buffer[i] in ['0'..'9', comma, point]) then
      begin
        buffer := '0' + spDecimal;
        for j := 1 to FPrecision do
          buffer := buffer + '0';
        valid := False;
        Break;
      end;

    //se o texto no tem caracteres invlidos,
    if valid then
    begin
      //apaga todos os pontos e vrgulas
      while Pos(comma, buffer) > 0 do
        Delete(buffer, Pos(comma, buffer), 1);
      while Pos(point, buffer) > 0 do
        Delete(buffer, Pos(point, buffer), 1);

      //coloca o separador decimal, se precisar
      case FAutoFormatDecimal of
        //se a formatao do separador decimal for automtica,
        True:
          begin //coloca o separador decimal na posio de acordo com a propriedade EditPrecision
            if (Length(buffer) > FPrecision) then
              Insert(spDecimal, buffer, (Length(buffer) - FPrecision + 1))
            else
            begin
              while (Length(buffer) < FPrecision) do buffer := '0' + buffer;
                buffer := '0' + spDecimal + buffer;
            end;
          end;
        //se a formatao do separador decimal for manual,
        False:
          begin //coloca o separador decimal seguido de zeros
            buffer := buffer + spDecimal;
            for i := 1 to FPrecision do
              buffer := buffer + '0';
          end;
      end;

      //busca a posio do primeiro separador de milhar
      i := (Pos(spDecimal, buffer) - 1) mod FThousandCount;
      if i = 0 then
        i := (FThousandCount + 1)
      else
        i := i + 1;

      //faz um loop para colocar um separador de milhar de 3 em 3 posies, se precisar
      while i < (Pos(spDecimal, buffer) - 1) do
      begin
        Insert(spThousand, buffer, i);
        i := i + FThousandCount + 1;
      end;
    end;
  end;

  //formata o valor
  Self.text := buffer;
end;

procedure TExtendedEdit.ValidateDigit(var Key: Char);
var
  spPosition: word;
begin
  if Key in ['0'..'9', comma, point, back, minus] then
  begin
    case Key of
      //se o dgito for o backspace
      back: FChangeThousand := True;

      //se o dgito for uma vrgula
      comma:
        case FDecimal of
          //e se o separador decimal for a vrgula,
          spComma:
            begin
              //busca a posio do separador decimal
              spPosition := Pos(comma, Self.text);
              //e se o separador decimal j existe,
              if (spPosition > 0) then
                Key := #0 //ento ignora
              //ou se vai estourar o nmero de casas decimais,
              else if (Length(Self.text) - Self.SelStart) > FPrecision then
                Key := #0 //ento ignora
              else
                FChangeThousand := True;
            end;
          //e se o separador decimal for o ponto,
          spPoint:
            begin
              //busca a posio do separador decimal
              spPosition := Pos(point, Self.text);
              //e se o separador decimal j existe, e sua posio  anterior a posio do cursor,
              if (spPosition > 0) then
                Key := #0 //ento ignora
              //ou se vai estourar o nmero de casas decimais,
              else if (Length(Self.text) - Self.SelStart) > FPrecision then
                Key := #0 //ento ignora
              //troca o dgito pelo separador decimal
              else
              begin
                Key := point;
                FChangeThousand := True;
              end;
            end;
        end;

      //se o dgito  um ponto
      point:
        case FDecimal of
          //e se o separador decimal for a vrgula,
          spComma:
            begin
              spPosition := Pos(comma, Self.text); //busca a posio do separador decimal
              //e se o separador decimal j existe, e sua posio  anterior a posio do cursor,
              if (spPosition > 0) then
                Key := #0 //ento ignora
              //ou se vai estourar o nmero de casas decimais,
              else if (Length(Self.text) - Self.SelStart) > FPrecision then
                Key := #0 //ento ignora
              //troca o dgito pelo separador decimal
              else
              begin
                Key := comma;
                FChangeThousand := True;
              end;
            end;
          //e se o separador decimal for o ponto,
          spPoint:
            begin
              spPosition := Pos(point, Self.text); //busca a posio do separador decimal
              //e se o separador decimal j existe,
              if (spPosition > 0) then
                Key := #0 //ento ignora
              //ou se vai estourar o nmero de casas decimais,
              else if (Length(Self.text) - Self.SelStart) > FPrecision then
                Key := #0 //ento ignora
              else
                FChangeThousand := True;
            end;
        end;

      //se o dgito  um nmero
      '0'..'9', minus:
        begin
          //busca a posio do separador decimal
          case FDecimal of
            spComma: spPosition := Pos(comma, Self.text);
            spPoint: spPosition := Pos(point, Self.text);
          end;
          //se as casas decimais j esto completas,
          if (spPosition > 0) and (Self.SelStart >= spPosition) and (Length(Self.text) - spPosition = FPrecision) then
            Key := #0 //ento ignora
          else
            FChangeThousand := True;
        end;
    end;
  end
  else
  begin
    if not (Key in [#8, #13, #35, #36, #37, #39]) then
    begin
      Key := #0;
      FChangeThousand := False;
    end;
  end;
end;

procedure TExtendedEdit.FormatThousand;
var
  i: integer;
  spPosition, crPosition: word;
  spDecimal, spThousand: char;
  buffer_int, buffer_dec: String;
begin
  //busca a posio do cursor
  crPosition := Self.SelStart;

  //busca o separador decimal e de milhar, e a posio do separador decimal
  case FDecimal of
    spComma:
      begin
        spDecimal  := comma;
        spPosition := Pos(comma, Self.text);
        spThousand := point;
      end;
    spPoint:
      begin
        spDecimal  := point;
        spPosition := Pos(point, Self.text);
        spThousand := comma;
      end;
  end;

  //se no tem separador decimal, fica valendo o tamanho do nmero
  if (spPosition = 0) then
    spPosition := Length(Self.text);

  //guarda a parte inteira e a parte decimal do nmero
  if Pos(spDecimal, Self.text) > 0 then
  begin
    buffer_int := Copy(Self.text, 0, spPosition - 1);
    buffer_dec := Copy(Self.text, spPosition, Length(Self.text));
  end
  else
  begin
    buffer_int := Self.text;
    buffer_dec := '';
  end;

  //se o cursor est antes do separador decimal,
  if (crPosition <= spPosition) then
  begin
    //apaga do nmero todos os separadores de milhar
    while (Pos(spThousand, buffer_int) > 0) do
      Delete(buffer_int, Pos(spThousand, buffer_int), 1);

    //busca a posio do primeiro separador de milhar
    i := Length(buffer_int) mod FThousandCount;
    if i = 0 then
      i := (FThousandCount + 1)
    else
      i := i + 1;

    //faz um loop para colocar um separador de milhar de 3 em 3 posies
    while i < Length(buffer_int) do
    begin
      Insert(spThousand, buffer_int, i);
      i := i + FThousandCount + 1;
    end;

    //recalcula a posio do cursor caso tenha surgido mais um separador de milhar na formatao
    crPosition := crPosition + (Length(buffer_int + buffer_dec) - Length(Self.text));

    //ajusta o nmero caso a parte decimal ou a parte inteira ficar vazia
    if (buffer_int = '') and (buffer_dec <> '') then
      buffer_int := '0';

    //coloca o valor formatado no componente
    Self.text := buffer_int + buffer_dec;

    //mantm o cursor na posio correta, ou seja, a direita do ltimo dgito inserido
    Self.SelStart := crPosition;
  end;
end;

procedure TExtendedEdit.SetColorOnFocus(Value: TColor);
begin
  if not (csDesigning in ComponentState) and FFocused then
    Color := Value;
  FColorOnFocus := Value;
end;

procedure TExtendedEdit.SetColorOnNoFocus(Value: TColor);
begin
  Color := Value;
  FColorOnNoFocus := Value;
end;

procedure TExtendedEdit.SetFontOnFocus(Value: TFont);
begin
  if not (csDesigning in ComponentState) and FFocused then
    Font.Assign(Value);
  FFontOnFocus.Assign(Value);
end;

procedure TExtendedEdit.SetFontOnNoFocus(Value: TFont);
begin
  Font.Assign(Value);
  FFontOnNoFocus.Assign(Value);
end;

procedure TExtendedEdit.DoEnter;
begin
  if (FEditType = etFloat) and (MaxLength = 0) then
    MaxLength := 16;
  TextAtEnter := Text;

  if FWidthOnFocus > 0 then begin
    iWidth := Width;
    Width  := FWidthOnFocus;
  end;

  SetColorOnFocus(FColorOnFocus);
  SetFontOnFocus(FFontOnFocus);
  // if Assigned(FOnEnter) then FOnEnter(Self);
  inherited;
end;

procedure TExtendedEdit.DoExit;
var
  k : Integer;
begin
  if (FEditType in [etDate, etTime]) then
    CheckDateTime;

  if WidthOnFocus > 0 then
    Width := iWidth;

  {Alignment of the String, when have a form defined alignment}
  if (FEditType = etString) and (Length(Text) > 0) then {String}
  begin
    {Pe o primeiro caracter em maisculas}
    if FUpper then begin
      if Length(Text) = 1 then
        Text := AnsiUpperCase(Text);
      if Length(Text) > 1 then
        Text := AnsiUpperCase(Text[1]) + Copy(Text,2,Length(Text)-1);
    end;
  end;

  {Checks range max for Date input}
  if (FEditType = etDate) and
     ((Trunc(FMaxValue) <> 0) or (Trunc(FMinValue) <> 0)) and
     ((Trunc(FValDateTime) < Trunc(FMinValue)) or (Trunc(FValDateTime) > Trunc(FMaxValue))) then
  begin
    ShowMessage('Range Max. : ' + FormatDateTime(FDateTimeMask, FMinValue) + ' <-> ' + FormatDateTime(FDateTimeMask, FMaxValue));
    Self.SetFocus;
    Exit;
  end;

  {Checks range max for Time input}
  if (FEditType = etTime) and
     ((FMaxValue <> 0) or (FMinValue <> 0)) and
     ((FValDateTime < FMinValue) or (FValDateTime > FMaxValue)) then
  begin
    ShowMessage('Range Max. : ' + FormatDateTime(FDateTimeMask, FMinValue) + ' <-> ' + FormatDateTime(FDateTimeMask, FMaxValue));
    Self.SetFocus;
    Exit;
  end;

  {Checks range max for Integer input}
  if (FEditType = etInteger) and
     ((Trunc(FMaxValue) <> 0) or (Trunc(FMinValue) <> 0)) and
     ((FValInteger < Trunc(FMinValue)) or (FValInteger > Trunc(FMaxValue))) then
  begin
    ShowMessage('Range Max. : ' + IntToStr(Trunc(FMinValue)) + ' <-> ' + IntToStr(Trunc(FMaxValue)));
    Self.SetFocus;
    Exit;
  end;

  {Checks range max for Float input}
  if (FEditType = etFloat) and
     ((FMaxValue <> 0) or (FMinValue <> 0)) and
     ((FValFloat < FMinValue) or (FValFloat > FMaxValue)) then
  begin
    ShowMessage('Range Max. : ' + FloatToStr(FMinValue) + ' <-> ' + FloatToStr(FMaxValue));
    Self.SetFocus;
    Exit;
  end;

  {If it has been defined precision, gives format to the string}
  if (FEditType = etFloat) and ((FPrecision > 0) or FUseThousand) then
  begin
    if Length(Text) = 0 then
      Text := '0';
    FSDecimal  := DecimalSeparator;
    FSThousand := ThousandSeparator;
    if (Pos(FSDecimal, Text) = 0) or FUseThousand then begin
      if FUseThousand then
        Text := FloatToStrF(FValFloat, ffNumber, 18, FPrecision)
      else
        Text := FloatToStrF(FValFloat, ffFixed, 18, FPrecision);
      {Text := Text + FSDecimal;
      for k := 1 to FPrecision do
        Text := Text + '0';}
    end else begin
      if Length(Text) - Pos(FSDecimal, Text) > FPrecision then
        Text := Copy(Text, 1, Pos(FSDecimal, Text) + FPrecision);
      if Length(Text) - Pos(FSDecimal, Text) < FPrecision then
        for k := Length(Text) - Pos(FSDecimal, Text) + 1 to FPrecision do
          Text := Text + '0';
    end;
  end;

  SetColorOnNoFocus(FColorOnNoFocus);
  SetFontOnNoFocus(FFontOnNoFocus);
  // if Assigned(FOnExit) then FOnExit(Self);
  inherited;
end;

procedure TExtendedEdit.FontChanged(Sender: TObject);
begin
  Font.Assign(FFontOnNoFocus);
end;

procedure TExtendedEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    //formata o nmero caso seja teclado <<DELETE>>
    VK_DELETE: FChangeThousand := (FEditType in [etInteger, etFloat]);
  end;
  if (FExitOnReturn) and (Key = VK_RETURN) then
    PostMessage(Handle, WM_KEYDOWN, VK_TAB, 0)
  else if (Key = VK_RETURN) then
    PostMessage(Parent.Handle, WM_KEYDOWN, VK_RETURN, 0)
  else if (Key = VK_ESCAPE) and (Text = TextAtEnter) then
    PostMessage(Parent.Handle, WM_KEYDOWN, VK_ESCAPE, 0);
  //inherited;
  inherited KeyDown(Key, Shift);
end;

procedure TExtendedEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
  if FChangeThousand then
    FormatThousand;
  FChangeThousand := False;
  inherited KeyUp(Key, Shift);
end;

procedure TExtendedEdit.KeyPress(var Key: Char);
var
// {$IFDEF VER100}
  FEditTemp : TCustomForm; {For Delphi 3}
// {$ELSE}
//  FEditTemp : TForm;       {For Delphi 1 - 2}
// {$ENDIF}
  C         : String;
begin
  if Key = FKeyTab then
  begin
    FEditTemp := GetParentForm(Self); //Change declaration in "var" for Delphi version
    SendMessage(FEditTemp.Handle, WM_NEXTDLGCTL, 0, 0);
    Key := #0;
  end
  else if (Key = #27) then // Change in version 2.01 by Fernando Val
  begin                    // If ESC is pressed during edit
    if (Text <> TextAtEnter) then // If any change was made in component text
    begin
      Text := TextAtEnter; // All changes are cancelled
      Key  := #15; // Delete'the ESC-key press
    end
    else
      Key  := #15; // Delete'the ESC-key press
  end
  else
  begin
    if FEditType = etString then
    begin
      // To put the first char with capital letters upon beginning to write
      if FUpper then
      begin // Change in Version 1.03
        if Length(Text) = 0 then
        begin
          C   := AnsiUpperCase(Key);
          Key := C[1];
        end;
      end;
      // To try the alone chain when it is numerical, to avoid to introduce wrong
      // characters
    end
    else if FEditType in [etInteger, etFloat] then
    begin
      // Permitted characters when it is Integer
      if FEditType = etInteger then
      begin
        if ((Pos(minus, Text) > 0) or (Key = minus)) and (MaxLength = 0) then
          MaxLength := 11;
        if not (Key in ['0'..'9', minus, #8, #13, #35, #36, #37, #39]) then
          Key := #15;
      end;

      // Permitted characters when it is Float
      if FEditType = etFloat then begin
        if not (Key in ['0'..'9', comma, point, minus, #8, #13, #35, #36, #37, #39]) then
          Key := #15;
        // if (Key = comma) or (Key = point) then
        //  if (Pos(comma, Text) > 0) or (Pos(point, Text) > 0) then
        //    Key := #15
        //  else
        //    Key := DecimalSeparator;
      end;

      // To eliminate the introduction from spaces
      if Key = #32 then
        Key := #15;

      // To verify that alone is introduce a negative sign.
      if (Key = minus) and (Pos(minus, Text) > 0) and (SelLength <> Length(Text)) then
        Key := #15;

      ValidateDigit(Key);
    end;
  end;
  if Key <> #0 then
    inherited KeyPress(Key);

  if {(FExitOnReturn) and} (ord(Key) = VK_RETURN) then
  begin
    DoExit;
    DoEnter;
    Key := chr(0);
  end
  else
    inherited;
end;

procedure TExtendedEdit.Change; {New and change in version 1.02}
var
  i : Integer;
  C : String;
begin
  {To convert the chain if it is numerical,to return a value}
  if ((FEditType = etInteger) or (FEditType = etFloat)) and (Length(Text) > 0) then
  begin
    if FEditType = etInteger then
    begin
      for i := 1 to Length(Text) do
      begin
        if Text[i] in ['0'..'9', minus, plus] then
          C := C + Text[i];
      end;
      Text := C;
    end;

    if FEditType = etFloat then
    begin
      for i := 1 to Length(Text) do
      begin
        if Text[i] in ['0'..'9', comma, point, minus, plus] then
          C := C + Text[i];
      end;
      Text := C;
    end;

    if Length(Text) = 0 then
    begin
      if Assigned(FOnChange) then
        FOnChange(Self);
      Exit;
    end;

    try
      FValFloat   := 0;
      FValInteger := 0;

      {Eliminar caracteres no permitidos e mover o sinal - para o comeo para
       que no d erro de converso}
{      i := 1;
      while i <= Length(Text) do
        if not (Text[i] in ['0'..'9', comma, point, minus]) then
          Text := Copy(Text, 1, i-1) + Copy(Text, i+1, Length(Text) - i)
        else
          i := i + 1;
}
      {Se s tiver o sinal negativo, daira erro}
      if (Pos(minus, Text) = 1) and (Length(Text) = 1) then
        Exit;

      {Temporary variable to accomplish the conversion}
      TxtConvert := Text;

      {To put the negative sign to the beginning}
      if (FEditType <> etString) and (Pos(minus, TxtConvert) > 1) then
        if Length(TxtConvert) = Pos(minus, TxtConvert) then
          TxtConvert := minus + Copy(TxtConvert, 1, Pos(minus, TxtConvert) - 1)
        else
          TxtConvert := minus + Copy(TxtConvert, 1, Pos(minus, TxtConvert) - 1) +
                        Copy(TxtConvert, Pos(minus, TxtConvert) + 1, Length(TxtConvert) - Pos(minus,TxtConvert));

      {Remove thousand separators of text to be converted}
      if FUseThousand then
        while Pos(FSThousand, TxtConvert) > 0 do
          Delete(TxtConvert, Pos(FSThousand, TxtConvert), 1);

      {Convert text to inger value if edit type is integer}
      if FEditType = etInteger then
      begin
        {Range control of Integer}
        ValTemp := StrToFloat(TxtConvert);
        if (ValTemp > 2147483647) or (ValTemp < -2147483647) then
        begin
          ShowMessage('Range Max. : -2147483647 <-> 2147483647');
          FValInteger := 0;
        end
        else
          FValInteger := StrToInt(TxtConvert);
      end;

      {El tipo Float - Double, permite valores hasta 5.0 * 10e-324 .. 1.7 * 10e308
      con 15-16 digitos significativos, por lo que solamente controlamos que el total
      no pase de 16 digitos. Hasta la fecha no he experimentado con valores Float tan
      altos, por lo que no me atrevo a condicionar algo que no conozco con exactitud.}
      if FEditType = etFloat then
        FValFloat := StrToFloat(TxtConvert);
    except
      on EConvertError do
      begin
        ShowMessage('Range Max. :' + #13 +
                    ' - Integer : -2147483647 <-> 2147483647' + #13 +
                    ' - Float   : 5.0e-324 <-> 1.7e+308');
        FValInteger := 0;
        FValFloat   := 0;
      end;
    end;
  end;

  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TExtendedEdit.SetEditType(Value: TEditTypes);
begin
  if Value <> FEditType then
  begin
    if Value in [etInteger, etFloat] then
      Alignment := taRightJustify
    else
      Alignment := taLeftJustify;
    FEditType := Value;
    EditMask  := '';
    if Value = etInteger then
      SetInteger(FValInteger)
    else if Value = etFloat then
      SetFloat(FValFloat)
    else if Value in [etDate, etTime] then
    begin
      SetDateEditMask;
      SetDateTime(FValDateTime);
    end;
  end;
end;

procedure TExtendedEdit.SetInteger(Value: Integer);
begin
  if FEditType = etInteger then
    if FUseThousand then
      Text := FloatToStrF(Value, ffNumber, 18, 0)
    else
      Text := IntToStr(Value);
end;

procedure TExtendedEdit.SetFloat(Value: Double);
begin
  if FEditType = etFloat then
    if FUseThousand then
      Text := FloatToStrF(Value, ffNumber, 18, FPrecision)
    else
      Text := FloatToStrF(Value, ffFixed, 18, FPrecision);
end;

procedure TExtendedEdit.SetDateTime(Value: TDateTime);
begin
  if FEditType in [etDate, etTime] then
    if Value = EmptyDate then
      Text := ''
    else
      Text := FormatDateTime(FDateTimeMask, Value);
end;

procedure TExtendedEdit.SetPrecision(Value: Integer);
begin
  if Value >= 0 then
  begin
    FPrecision := Value;
    if FEditType = etFloat then
      SetFloat(FValFloat);
  end;
end;

procedure TExtendedEdit.SetDateOrder(Value: TDateOrder);
begin
  if Value <> FDateOrder then
  begin
    FDateOrder := Value;
    SetDateEditMask;
  end;
end;

procedure TExtendedEdit.SetTimeFormat(Value: TTimeFormat);
begin
  if Value <> FTimeFormat then
  begin
    FTimeFormat := Value;
    SetDateEditMask;
  end;
end;

procedure TExtendedEdit.SetDateEditMask;
begin
  if FEditType = etDate then
  begin
    case FDateOrder of
      doDMY : FDateTimeMask := 'DD/MM/YYYY';
      doMDY : FDateTimeMask := 'MM/DD/YYYY';
      doYMD : FDateTimeMask := 'YYYY/MM/DD';
    end;

    case FDateOrder of
      doDMY, doMDY : EditMask := '99/99/9999;1; ';
      doYMD : EditMask := '9999/99/99;1; ';
    end;
  end
  else if FEditType = etTime then
  begin
    case FTimeFormat of
      tfHMS : FDateTimeMask := 'HH:NN:SS';
      tfHM  : FDateTimeMask := 'HH:NN';
      tfMS  : FDateTimeMask := 'NN:SS';
    end;

    case FTimeFormat of
      tfHMS : EditMask := '99:99:99;1; ';
      tfHM, tfMS : EditMask := '99:99;1; ';
    end;
  end;
end;

procedure TExtendedEdit.CheckDateTime;
var
  wDia, wMes, wAno : Word;
  sAno : string;
begin
  if FEditType = etDate then
  begin
    if Trim(Text) <> '/  /' then
      try
        wDia := StrToInt(Copy(Text, Pos('DD', FDateTimeMask), 2));
        wMes := StrToInt(Copy(Text, Pos('MM', FDateTimeMask), 2));
        sAno := Copy(Text, Pos('YYYY', FDateTimeMask), 4);
        if Copy(sAno, 3, 2) = '  ' then
          wAno := StrToInt(Copy(sAno, 1, 2))
        else
          wAno := StrToInt(sAno);
        FValDateTime := EncodeDate(wAno, wMes, wDia);
        SetDateTime(FValDateTime);
      except
        on EConvertError do
        begin
          FValDateTime := EmptyDate;
          if FCheckOnExit then
          begin
            Self.SetFocus;
            raise EConvertError.Create(StringReplace(FMsgInvDate, '%text%', Text, [rfReplaceAll, rfIgnoreCase]));
          end;
        end;
      end;
  end
  else if FEditType = etTime then
  begin
    if Trim(Text) <> ':  :' then
      try
        case FTimeFormat of
          tfHMS : FValDateTime := StrToTime(Text);
          tfHM  : FValDateTime := StrToTime(Text + ':00');
          tfMS  : FValDateTime := StrToTime('00:' + Text);
        end;
      except
        on EConvertError do
        begin
          FValDateTime := EmptyDate;
          if FCheckOnExit then
          begin
            Self.SetFocus;
            raise EConvertError.Create(StringReplace(FMsgInvTime, '%text%', Text, [rfReplaceAll, rfIgnoreCase]));
          end;
        end;
      end;
  end;
end;

procedure register;
begin
  RegisterComponents('Additional', [TExtendedEdit]);
end;

end.
